use json;

/* Read instance data */
function input() {
    local usage = "Usage: localsolver pooling.lsp "
            + "inFileName=inputFile [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;

    problem = json.parse(inFileName);

    nbComponents = problem.components.count();
    nbProducts =  problem.products.count();
    nbPools = problem.pool_size.count();
    nbAttributes = problem.components[0].quality.count();

    // Components
    componentQualities[c in 0...nbComponents] = problem.components[c].quality.values();
    for [c in 0...nbComponents]
        componentIdx[problem.components[c].name] = c;
    
    // Final products (blendings)
    for [p in 0...nbProducts]
        productIdx[problem.products[p].name] = p;
    for [p in 0...nbProducts] {
        if (problem.products[p]["quality_lower"] == nil)
            minTolerance[p][k in 0...nbAttributes] = 0;
        else
            minTolerance[p] = problem.products[p].quality_lower.values();
    }
    maxTolerance[p in 0...nbProducts] = problem.products[p].quality_upper.values();

    // Intermediate pools
    poolCapacities = problem.pool_size.values();
    poolNames = problem.pool_size.keys();
    for [o in 0...nbPools]
        poolIdx[poolNames[o]] = o;

    // Flow graph

    // Edges from the components to the products
    upperBoundComponentToProduct[c in 0...nbComponents][p in 0...nbProducts] = 0;
    costComponentToProduct[c in 0...nbComponents][p in 0...nbProducts] = 0;

    // Edges from the components to the pools
    upperBoundFractionComponentToPool[c in 0...nbComponents][o in 0...nbPools] = 0;
    costComponentToPool[c in 0...nbComponents][o in 0...nbPools] = 0;
    
    // Edges from the pools to the products
    upperBoundPoolToProduct[o in 0...nbPools][p in 0...nbProducts] = 0;
    costPoolToProduct[o in 0...nbPools][p in 0...nbProducts] = 0;

    // Bound and cost on the edges
    for [edge in problem.component_to_product_bound] {
        upperBoundComponentToProduct[componentIdx[edge.component]]
                [productIdx[edge.product]] = edge.bound;
        if (count(edge) > 3) 
            costComponentToProduct[componentIdx[edge.component]]
                    [productIdx[edge.product]] = edge.cost;
    }

    for [edge in problem.component_to_pool_fraction] {
        upperBoundFractionComponentToPool[componentIdx[edge.component]]
                [poolIdx[edge.pool]] = edge.fraction;
        if (count(edge) > 3) 
            costComponentToPool[componentIdx[edge.component]]
                    [poolIdx[edge.pool]] = edge.cost;
    }

    for [edge in problem.pool_to_product_bound] {
        upperBoundPoolToProduct[poolIdx[edge.pool]]
                [productIdx[edge.product]] = edge.bound;
        if (count(edge) > 3) 
            costPoolToProduct[poolIdx[edge.pool]]
                    [productIdx[edge.product]] = edge.cost;
    }
}

/* Declare the optimization model */
function model() {

    // Decision variables

    // Flow from the components to the products
    flowComponentToProduct[c in 0...nbComponents][p in 0...nbProducts] <-
            float(0, upperBoundComponentToProduct[c][p]);

    // Fraction of the total flow in pool o coming from the component c
    fractionComponentToPool[c in 0...nbComponents][o in 0...nbPools] <-
            float(0, upperBoundFractionComponentToPool[c][o]);

    // Flow from the pools to the products
    flowPoolToProduct[o in 0...nbPools][p in 0...nbProducts] <-
            float(0, upperBoundPoolToProduct[o][p]);

    // Flow from the components to the products and passing by the pools
    flowComponentToProductByPool[c in 0...nbComponents][o in 0...nbPools][p in 0...nbProducts] <-
            fractionComponentToPool[c][o] * flowPoolToProduct[o][p];

    // Constraints

    // Proportion
    for [o in 0...nbPools] {
       proportion <- sum[c in 0...nbComponents](fractionComponentToPool[c][o]);
       constraint proportion == 1;
    }

    // Component supply
    for [c in 0...nbComponents] {
        flowToProducts <- sum[p in 0...nbProducts](flowComponentToProduct[c][p]);
        flowToPools <- sum[o in 0...nbPools][p in 0...nbProducts]
                (flowComponentToProductByPool[c][o][p]);
        totalOutFlow <- flowToPools + flowToProducts;
        constraint totalOutFlow <= problem.components[c].upper;
    }

    // Pool capacity (bounds on edges)
    for [c in 0...nbComponents] {
        for [o in 0...nbPools] {
            flowComponentToPool <- sum[p in 0...nbProducts]
                    (flowComponentToProductByPool[c][o][p]);
            edgeCapacity <- poolCapacities[o] * fractionComponentToPool[c][o];
            constraint flowComponentToPool <= edgeCapacity;
        }
    }

    // Product capacity
    for [p in 0...nbProducts] {
        flowFromPools <- sum[o in 0...nbPools](flowPoolToProduct[o][p]);
        flowFromComponents <- sum[c in 0...nbComponents](flowComponentToProduct[c][p]);
        totalInFlow <- flowFromComponents + flowFromPools;
        constraint totalInFlow <= problem.products[p].upper;
        constraint totalInFlow >= problem.products[p].lower;
    }

    // Product tolerance 
    for [p in 0...nbProducts] {
        for [k in 0...nbAttributes] {
            
            // Attribute from the components
            attributeFromComponents <- sum[c in 0...nbComponents]
                    (componentQualities[c][k] * flowComponentToProduct[c][p]);
            
            // Attribute from the pools
            attributeFromPools <- sum[c in 0...nbComponents](componentQualities[c][k] *
                    sum[o in 0...nbPools](flowComponentToProductByPool[c][o][p]));
            
            // Total flow in the blending
            totalFlowIn <- sum[o in 0...nbPools](flowPoolToProduct[o][p]) +
                    sum[c in 0...nbComponents](flowComponentToProduct[c][p]);

            totalAttributeIn <- attributeFromComponents  + attributeFromPools;
            constraint totalAttributeIn >= minTolerance[p][k] * totalFlowIn;
            constraint totalAttributeIn <= maxTolerance[p][k] * totalFlowIn;
        }
    }

    // Objective function

    // Cost of the flows from the components directly to the products
    directFlowCost <- sum[c in 0...nbComponents][p in 0...nbProducts]
            (flowComponentToProduct[c][p] * costComponentToProduct[c][p]);

    // Cost of the flows from the components to the products passing by the pools
    undirectFlowCost <- sum[c in 0...nbComponents][o in 0...nbPools][p in 0...nbProducts]
            ((costComponentToPool[c][o] + costPoolToProduct[o][p]) *
            flowComponentToProductByPool[c][o][p]);

    // Gain of selling the final products
    productsGain <- sum[p in 0...nbProducts](problem.products[p].price *
            (sum[o in 0...nbPools](flowPoolToProduct[o][p]) +
            sum[c in 0...nbComponents](flowComponentToProduct[c][p])));
    
    // Cost of buying the components
    componentsCost <- sum[c in 0...nbComponents](problem.components[c].price *
            (sum[o in 0...nbPools][p in 0...nbProducts]
            (fractionComponentToPool[c][o] * flowPoolToProduct[o][p]) +
            sum[p in 0...nbProducts](flowComponentToProduct[c][p])));

    profit <- productsGain - componentsCost - (directFlowCost + undirectFlowCost);
    
    // Maximize the total profit
    maximize profit;
}

/* Parameterize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file */
function output() {
    if (solFileName == nil) return;

    // Solution flow from the components to the products
    for [c in 0...nbComponents] {
        for [p in 0...nbProducts] {
            component_to_poduct[c * nbProducts + p] = {
                    "component" = problem.components[c].name,
                    "product" = problem.products[p].name,
                    "flow" = flowComponentToProduct[c][p].value};
        }
    }
    // Solution fraction of the inflow at pool o coming from the component c
    for [c in 0...nbComponents] {
        for [o in 0...nbPools] {
            component_to_pool_fraction[c * nbPools + o] = {
                    "component" = problem.components[c].name,
                    "pool" = poolNames[o],
                    "flow" = fractionComponentToPool[c][o].value};
        }
    }
    // Solution flows from the pools to the products
    for [o in 0...nbPools] {
        for [p in 0...nbProducts] {
            pool_to_product[o * nbProducts + p] = {
                    "pool" = poolNames[o],
                    "product" = problem.products[p].name,
                    "flow" = flowPoolToProduct[o][p].value};
        }
    }

    json.dump({"objective" = profit.value, "solution" = {
            "component_to_pool_fraction" = component_to_pool_fraction,
            "component_to_product" = component_to_poduct ,
            "pool_to_product" = pool_to_product}}, solFileName);
}
